@@ -5,6 +5,11 @@ from typing import TYPE_CHECKING, Union, Tuple, Sequence
 from numpy.typing import ArrayLike
 if TYPE_CHECKING:
     from neurograd.tensor import Tensor
+try:
+    import opt_einsum
+except ImportError:
+    opt_einsum = None
+
 
 def _transpose(arr):
     """Transpose that works for any ndim >= 1"""
@@ -42,7 +47,9 @@ class Linear(Function, Module):
         Function.__init__(self)
         Module.__init__(self)
     def forward(self, X: xp.ndarray, W: xp.ndarray, b: xp.ndarray) -> xp.ndarray:
-        return xp.matmul(X, W) + b
+        out = xp.matmul(X, W)
+        xp.add(out, b, out=out)
+        return out
     def backward(self, grad_output: xp.ndarray) -> xp.ndarray:
         X, W, b = self.parent_tensors
         grad_X = xp.matmul(grad_output, _transpose(W.data)) if X.requires_grad else None
@@ -108,15 +115,28 @@ class TensorDot(Function, Module):
 
 class EinSum(Function, Module):
     name = "EinSum"
-    def __init__(self, subscripts: str):
+    def __init__(self, subscripts: str, optimize=False):
         Function.__init__(self)
         Module.__init__(self)
         self.subscripts = subscripts.replace(" ", "")
+        self.optimize = optimize
+        self.forward_path = None
+        self.backward_paths = {}
     def forward(self, *operands: xp.ndarray) -> xp.ndarray:
         self.operand_shapes = [op.shape for op in operands]
         self.input_operands = operands
-        return xp.einsum(self.subscripts, *operands)   
+        self.parent_tensors = operands 
+        if opt_einsum is not None and self.optimize:
+            if self.forward_path is None:
+                dummy_arrays = [xp.zeros(shape) for shape in self.operand_shapes]
+                self.forward_path = opt_einsum.contract_path(self.subscripts, *dummy_arrays, optimize='auto')
+            return opt_einsum.contract(self.subscripts, *operands, 
+                                     backend='cupy' if xp.__name__ == 'cupy' else 'numpy',
+                                     optimize=self.forward_path)
+        return xp.einsum(self.subscripts, *operands, optimize=self.optimize)
     def backward(self, grad_output: xp.ndarray):
+        if not hasattr(self, 'parent_tensors'):
+            return tuple()
         grads = []
         if '->' in self.subscripts:
             inputs_str, output_str = self.subscripts.split('->')
@@ -124,26 +144,30 @@ class EinSum(Function, Module):
         else:
             input_specs = self.subscripts.split(',')
             all_indices = ''.join(input_specs)
-            output_str = ''.join(sorted(set(all_indices) - 
-                                      set([idx for idx in all_indices 
-                                           if all_indices.count(idx) > 1])))
+            output_str = ''.join([idx for idx in set(all_indices) if all_indices.count(idx) == 1])
         for i, tensor in enumerate(self.parent_tensors):
-            if not tensor.requires_grad:
+            if hasattr(tensor, 'requires_grad') and not tensor.requires_grad:
                 grads.append(None)
                 continue
             grad_operands = [grad_output]
-            grad_specs = [output_str]
+            grad_specs = [output_str] 
             for j, operand in enumerate(self.parent_tensors):
                 if j != i:
-                    grad_operands.append(operand.data)
+                    grad_operands.append(getattr(operand, 'data', operand))
                     grad_specs.append(input_specs[j])
             grad_equation = ','.join(grad_specs) + '->' + input_specs[i]
-            try:
-                grad = xp.einsum(grad_equation, *grad_operands)
-                grad = self._handle_broadcasting(grad, self.operand_shapes[i])
-                grads.append(grad)
-            except Exception as e:
-                grads.append(xp.zeros_like(self.input_operands[i]))
+            if opt_einsum is not None and self.optimize:
+                if grad_equation not in self.backward_paths:
+                    dummy_shapes = [grad_output.shape] + [getattr(op, 'data', op).shape 
+                                                         for j, op in enumerate(self.parent_tensors) if j != i]
+                    dummy_arrays = [xp.zeros(shape) for shape in dummy_shapes]
+                    self.backward_paths[grad_equation] = opt_einsum.contract_path(grad_equation, *dummy_arrays, optimize='auto')
+                grad = opt_einsum.contract(grad_equation, *grad_operands,
+                                         backend='cupy' if xp.__name__ == 'cupy' else 'numpy',
+                                         optimize=self.backward_paths[grad_equation])
+            else:
+                grad = xp.einsum(grad_equation, *grad_operands, optimize=self.optimize)   
+            grads.append(self._handle_broadcasting(grad, self.operand_shapes[i]))
         return tuple(grads)
     def _handle_broadcasting(self, grad, original_shape):
         if grad.shape == original_shape:
@@ -187,7 +211,7 @@ def dot(A, B):
     return MatMul()(A, B)
 def tensordot(A, B, axes):
     return TensorDot(axes)(A, B)
-def einsum(subscripts: str, *operands: xp.ndarray) -> xp.ndarray:
-    return EinSum(subscripts)(*operands)
+def einsum(subscripts: str, *operands: xp.ndarray, optimize = False) -> xp.ndarray:
+    return EinSum(subscripts, optimize=optimize)(*operands)
 def transpose(A, axes=None):
-    return Transpose(axes)(A)
+    return Transpose(axes)(A)
\ No newline at end of file
